@@ -12,8 +12,12 @@ import torch.nn.functional as F
 from siren_pytorch import SirenNet, SirenWrapper
 from torch import nn
 from torch.cuda.amp import GradScaler, autocast
-from torch.optim import Adam
+from torch_optimizer import DiffGrad, AdamP
+
+from PIL import Image
+import torchvision.transforms as T
 from torchvision.utils import save_image
+
 from tqdm import trange, tqdm
 
 from deep_daze.clip import load, tokenize
@@ -41,6 +45,9 @@ def exists(val):
     return val is not None
 
 
+def default(val, d):
+    return val if exists(val) else d
+
 def interpolate(image, size):
     return F.interpolate(image, (size, size), mode='bilinear', align_corners=False)
 
@@ -90,6 +97,8 @@ class DeepDaze(nn.Module):
             num_layers=8,
             image_width=512,
             loss_coef=100,
+            theta_initial=None,
+            theta_hidden=None
     ):
         super().__init__()
         # load clip
@@ -101,12 +110,17 @@ class DeepDaze(nn.Module):
         self.total_batches = total_batches
         self.num_batches_processed = 0
 
+        w0 = default(theta_hidden, 30.)
+        w0_initial = default(theta_initial, 30.)
+
         siren = SirenNet(
             dim_in=2,
             dim_hidden=256,
             num_layers=num_layers,
             dim_out=3,
-            use_bias=True
+            use_bias=True,
+            w0=w0,
+            w0_initial=w0_initial
         )
 
         self.model = SirenWrapper(
@@ -206,7 +220,11 @@ class Imagine(nn.Module):
             save_progress=False,
             seed=None,
             open_folder=True,
-            save_date_time=False
+            save_date_time=False,
+            start_image_path=None,
+            start_image_train_iters=10,
+            theta_initial=None,
+            theta_hidden=None
     ):
 
         super().__init__()
@@ -226,12 +244,14 @@ class Imagine(nn.Module):
             total_batches=total_batches,
             batch_size=batch_size,
             image_width=image_width,
-            num_layers=num_layers
+            num_layers=num_layers,
+            theta_initial=theta_initial,
+            theta_hidden=theta_hidden
         ).cuda()
 
         self.model = model
         self.scaler = GradScaler()
-        self.optimizer = Adam(model.parameters(), lr)
+        self.optimizer = AdamP(model.parameters(), lr)
         self.gradient_accumulate_every = gradient_accumulate_every
         self.save_every = save_every
         self.save_date_time = save_date_time
@@ -242,7 +262,24 @@ class Imagine(nn.Module):
         self.filename = self.image_output_path()
         self.encoded_text = tokenize(text).cuda()
 
-    def image_output_path(self, current_iteration=None):
+        self.start_image = None
+        self.start_image_train_iters = start_image_train_iters
+        if exists(start_image_path):
+            file = Path(start_image_path)
+            assert file.exists(), f'file does not exist at given starting image path {self.start_image_path}'
+            image = Image.open(str(file))
+
+            transform = T.Compose([
+                T.Resize(image_width),
+                T.CenterCrop((image_width, image_width)),
+                T.ToTensor(),
+                T.Normalize(0.5, 0.5)
+            ])
+
+            image_tensor = transform(image)[None, ...].cuda()
+            self.start_image = image_tensor
+
+    def image_output_path(self, sequence_number=None):
         """
         Returns underscore separated Path.
         A current timestamp is prepended if `self.save_date_time` is set.
@@ -250,8 +287,7 @@ class Imagine(nn.Module):
         :rtype: Path
         """
         output_path = self.textpath
-        if current_iteration:
-            sequence_number = int(current_iteration / self.save_every)
+        if sequence_number:
             sequence_number_left_padded = str(sequence_number).zfill(6)
             output_path = f"{output_path}.{sequence_number_left_padded}"
         if self.save_date_time:
@@ -259,16 +295,15 @@ class Imagine(nn.Module):
             output_path = f"{current_time}_{output_path}"
         return Path(f"{output_path}.png")
 
-
-    def generate_and_save_image(self, current_iteration=None):
+    def generate_and_save_image(self, sequence_number=None):
         """
-        :param current_iteration:
+        :param sequence_number:
         :param custom_filename: A custom filename to use when saving - e.g. "testing.png"
         """
         with torch.no_grad():
             img = normalize_image(self.model(self.encoded_text, return_loss=False).cpu())
             img.clamp_(0., 1.)
-            self.filename = self.image_output_path(current_iteration=current_iteration)
+            self.filename = self.image_output_path(sequence_number=sequence_number)
             save_image(img, self.filename)
             save_image(img, f"{self.textpath}.png")
 
@@ -289,11 +324,32 @@ class Imagine(nn.Module):
         self.optimizer.zero_grad()
 
         if (iteration % self.save_every == 0) and self.save_progress:
-            self.generate_and_save_image(current_iteration=iteration)
+            current_total_iterations = epoch * self.iterations + iteration
+            sequence_number = current_total_iterations // self.save_every
+            self.generate_and_save_image(sequence_number=sequence_number)
 
         return total_loss
 
     def forward(self):
+        if exists(self.start_image):
+            tqdm.write('Preparing with initial image...')
+            optim = DiffGrad(self.model.parameters(), lr = 3e-4)
+            pbar = trange(self.start_image_train_iters, desc='iteration')
+            for _ in pbar:
+                loss = self.model.model(self.start_image)
+                loss.backward()
+                pbar.set_description(f'loss: {loss.item():.2f}')
+
+                optim.step()
+                optim.zero_grad()
+
+                if terminate:
+                    print('interrupted by keyboard, gracefully exiting')
+                    return exit()
+
+            del self.start_image
+            del optim
+
         tqdm.write(f'Imagining "{self.text}" from the depths of my weights...')
 
         if self.open_folder:
